/**
	$Id: AgentController.java 6444 2014-04-09 12:21:20Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.jobagent;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.ConnectException;
import java.net.SocketTimeoutException;
import java.util.Properties;

import net.sf.basedb.clients.jobagent.handlers.DefaultRequestHandler;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.JobAgent;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.util.SocketUtil;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.jobagent.JobAgentConnection;
import net.sf.basedb.util.jobagent.JobAgentInfo;

/**
	This is the command line controller class for the job agent application. It is 
	responsible for starting up and managing a running job agent. The agent itself will 
	be started in separate thread.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2014-04-09 14:21:20 +0200 (on, 09 apr 2014) $
*/
public class AgentController
{

	/**
		Log job agent events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(AgentController.class);

	public static void main(String[] args)
	{
		int exitCode = 0;
		CmdLine cmdLine = new CmdLine(args);
		String cmd = cmdLine.getCmd();
		String propertiesFile = cmdLine.getOption("-c", "jobagent.properties");
		
		if ("help".equals(cmd))
		{
			printUsage();
			printHelp();
		}
		else
		{
			Properties p = new Properties();
			try
			{
				p.load(new FileInputStream(new File(propertiesFile)));
			}
			catch (Throwable t)
			{
				String message = "Could not load configuration file " + propertiesFile;
				System.out.println(message + ": " + t.getMessage());
				printUsage();
				log.debug(message, t);
				System.exit(1);
			}
		
			try
			{
				AgentController ac = new AgentController(p);
				if ("start".equals(cmd))
				{
					System.out.println("Starting job agent...");
					ac.startAgent();
					exitCode = -1; // Do not exit
					System.out.println("Job agent is up and running.");
				}
				else if ("stop".equals(cmd))
				{
					System.out.println("Stopping job agent...");
					ac.stopAgent();
					System.out.println("The job agent has been stopped");
				}
				else if ("info".equals(cmd))
				{
					System.out.println("Getting info from job agent...");
					JobAgentInfo info = ac.getInfo();
					System.out.println(info.toString());
				}
				else if ("status".equals(cmd))
				{
					String msg = "";
					try
					{
						JobAgentInfo info = ac.getStatus();
						msg += info.isPaused() ? "Paused" : "Running";
					}
					catch (ConnectException ex)
					{
						// ConnectException is expected if the agent is not running
						msg += "Stopped";
						log.debug(ex.getMessage(), ex);
					}
					catch (Throwable t)
					{
						msg += "Error (" + t.getMessage() + ")";
						log.error(msg, t);
						exitCode = 1;
					}
					System.out.println(msg);
				}
				else if ("register".equals(cmd))
				{
					System.out.println("Registering job agent with BASE...");
					ac.registerAgent();
					System.out.println("Done");
				}
				else if ("unregister".equals(cmd))
				{
					System.out.println("Unregistering job agent from BASE...");
					ac.unregisterAgent();
					System.out.println("Done");
				}
				else if ("pause".equals(cmd))
				{
					System.out.println("Pausing job agent...");
					ac.pauseAgent();
					System.out.println("The job agent is now paused");
				}
				else
				{
					System.out.println("Unknown command: " + cmd);
					printUsage();
					printHelp();
				}
			}
			catch (Throwable t)
			{
				exitCode = 1;
				String message = "Exception when executing command " + cmd;
				System.out.println(message + ": " + t.getMessage());
				log.error(message, t);
			}
		}
		if (exitCode >= 0) System.exit(exitCode);
	}

	private static void printUsage()
	{
		System.out.println("Usage    : jobagent.sh [options] cmd");
		System.out.println(" options : -c <configuration file>");
		System.out.println(" cmd     : start | stop | pause | info | status | register | unregister | help");
	}

	private static void printHelp()
	{
		System.out.println("Commands");
		System.out.println("===========");
		System.out.println("start     : Starts the job agent");
		System.out.println("stop      : Stops the job agent");
		System.out.println("pause     : Pauses the job agent");
		System.out.println("info      : Get info about the job agent");
		System.out.println("status    : Get status information about the job agent");
		System.out.println("register  : Registers this job agent with the BASE server");
		System.out.println("unregister: Unregisters this job agent from the BASE server");
		System.out.println("help      : Display this help message");
	}
	
	private final Properties properties;
	private final int port;
	private final int timeout;
	
	/**
		Create a new controller for controlling the job agent specified by the
		given properties.
		@param p The properties for the job agent to control
	*/
	public AgentController(Properties p)
	{
		this.properties = p;
		this.port = Values.getInt(p.getProperty("agent.port"), JobAgent.DEFAULT_PORT);
		this.timeout = Values.getInt(p.getProperty("agent.timeout"), 1000);
	}
	
	/**
		Start the agent this controller is controlling. If no agent is running
		a new one is created in another thread in this virtual machine. The thread
		must be stopped before the virtual machine exits.
		
		@throws IOException If there is an error
		@see JobAgentConnection#sendStart()
	*/
	public void startAgent()
		throws IOException
	{
		log.info("Starting job agent on port " + port);
		JobAgentConnection conn = new JobAgentConnection(port, timeout);
		try
		{
			JobAgentInfo info = null;
			try
			{
				log.info("Sending info command to job agent on port " + port);
				info = conn.getInfo(false);
			}
			catch (ConnectException t)
			{
				// Connection was refused, probably due to nobody listening on the port
				log.debug(t.getMessage(), t);
			}
			catch (SocketTimeoutException t)
			{
				// Connection timed out, maybe because nobody listens
				log.debug(t.getMessage(), t);
			}
			
			if (info == null)
			{
				// No job agent is running, create a new agent
				log.info("Creating a new job agent on port " + port);
				Agent agent = createAgent(properties);
				agent.service(new DefaultRequestHandler(agent)); // Start listening for incoming connections
			}
			else
			{
				log.info("An existing job agent is already running on port " + port + 
					" (" + (info.isPaused() ? "paused" : "running") + ")");
			}

			// If we reach this part, we should at least have a job agent listening for connections
			if (info == null || info.isPaused())
			{
				// It is a new agent or an existing that is paused --> Start it
				log.info("Sending start command to job agent on port " + port);
				conn.sendStart();
			}
		}
		finally
		{
			conn.close();
		}
	}

	/**
		Stop a running job agent by sending a stop request to the agents remote control
		service. The agent may be running in this or in another virtual machine.
		@throws IOException If there is an error
		@see JobAgentConnection#sendStop()
	*/
	public void stopAgent()
		throws IOException
	{
		log.info("Stopping job agent on port " + port);
		JobAgentConnection conn = new JobAgentConnection(port, timeout);
		try
		{
			log.info("Sending stop command to job agent on port " + port);
			conn.sendStop();
		}
		catch (ConnectException t)
		{
			// Connection was refused, probably due to nobody listening on the port
			// Do nothing
			log.debug(t.getMessage(), t);
			log.info("No job agent is running on port " + port);
		}
		finally
		{
			conn.close();
		}
	}
	
	/**
		Pause a running job agent by sending a stop request to the agents remote control
		service. The agent may be running in this or in another virtual machine.
		@throws IOException If there is an error
		@see JobAgentConnection#sendPause()
	*/
	public void pauseAgent()
		throws IOException
	{
		log.info("Pausing job agent on port " + port);
		JobAgentConnection conn = new JobAgentConnection(port, timeout);
		try
		{
			log.info("Sending stop command to job agent on port " + port);
			conn.sendPause();
		}
		catch (ConnectException t)
		{
			// Connection was refused, probably due to nobody listening on the port
			// Do nothing
			log.debug(t.getMessage(), t);
			log.info("No job agent is running on port " + port);
		}
		finally
		{
			conn.close();
		}
	}

	/**
		Get info about running job agent by sending an info request to the agents remote control
		service. The agent may be running in this or in another virtual machine.
		@throws IOException If there is an error
		@return A <code>JobAgentInfo</code> object
		@see JobAgentConnection#getInfo(boolean)
	*/
	public JobAgentInfo getInfo()
		throws IOException
	{
		log.info("Getting info from agent on port " + port);
		JobAgentConnection conn = new JobAgentConnection(port, timeout);
		JobAgentInfo info = null;
		try
		{
			log.info("Sending info command to job agent on port " + port);
			info = conn.getInfo(true);
		}
		finally
		{
			conn.close();
		}
		return info;
	}
	
	/**
		Get status info about running job agent by sending an info request to the agents r
		emote control service. The agent may be running in this or in another virtual machine.
		@throws IOException If there is an error
		@return A <code>JobAgentInfo</code> object
		@see JobAgentConnection#getInfo(boolean)
		@since 2.5
	*/
	public JobAgentInfo getStatus()
		throws IOException
	{
		log.info("Getting status from agent on port " + port);
		JobAgentConnection conn = new JobAgentConnection(port, timeout);
		JobAgentInfo info = null;
		try
		{
			log.info("Sending status command to job agent on port " + port);
			info = conn.getInfo(true);
		}
		finally
		{
			conn.close();
		}
		return info;
	}

	public void registerAgent()
	{
		Agent agent = createAgent(properties);
		log.info("Registering agent '" + agent.getId() + "' with BASE");
		SessionControl sc = agent.getSessionControl();
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			JobAgent jobAgent = null;
			try
			{
				jobAgent = agent.getJobAgent(dc);
				log.info("Agent with id '" + agent.getId() + "' is already registered");
			}
			catch (ItemNotFoundException ex)
			{
				jobAgent = JobAgent.getNew(dc, agent.getId());
				jobAgent.setName(agent.getName());
				jobAgent.setDescription(agent.getDescription());
				jobAgent.setPort(agent.getPort());
				jobAgent.setServer(SocketUtil.getPublicLocalHost().getHostName());
				dc.saveItem(jobAgent);
				dc.commit();
			}
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}

	public void unregisterAgent()
	{
		Agent agent = createAgent(properties);
		SessionControl sc = agent.getSessionControl();
		DbControl dc = null;
		try
		{
			dc = sc.newDbControl();
			JobAgent jobAgent = null;
			try
			{
				jobAgent = agent.getJobAgent(dc);
				dc.deleteItem(jobAgent);
				dc.commit();
			}
			catch (ItemNotFoundException ex)
			{}
		}
		finally
		{
			if (dc != null) dc.close();
		}
	}
	
	private Agent createAgent(Properties p)
	{
		Agent a = new Agent(p);
		return a;
	}
	
}
