/**
	$Id: BioPlate.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2008 Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.Transactional.Action;
import net.sf.basedb.core.data.BioPlateData;
import net.sf.basedb.core.data.BioWellData;
import net.sf.basedb.core.data.HardwareData;
import net.sf.basedb.core.data.PlateCoordinate;
import net.sf.basedb.core.data.PlateGeometryData;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

import java.util.Date;
import java.util.HashSet;
import java.util.Set;

/**
 	This class is represents physical bio-plates. A bio-plate must have
	a {@link PlateGeometry}
	<p>
	A bio-plate consists of {@link BioWell}:s which are organised into rows and
	columns as specified by the plate geometry.	
  	@author martin
  	@version 2.10
  	@base.modified $Date: 2017-03-13 09:15:46 +0100 (må, 13 mar 2017) $
 	
 */
public class BioPlate 
	extends AnnotatedItem
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#BIOPLATE
		@see #getType()
	*/
	public static final Item TYPE = Item.BIOPLATE;

	/**
		The maximum length of the barcode that can be stored in the
		database. Check the length against this value before calling the
		{@link #setBarcode(String)} method to avoid exceptions.
	*/
	public static final int MAX_BARCODE_LENGTH = BioPlateData.MAX_BARCODE_LENGTH;
	
	/**
		The maximum length of the section coordinate that can be stored in the
		database. Check the length against this value before calling the
		{@link #setSection(String)} method to avoid exceptions.
		@since 3.2
	*/
	public static final int MAX_SECTION_LENGTH = BioPlateData.MAX_SECTION_LENGTH;

	/**
		The maximum length of the tray coordinate that can be stored in the
		database. Check the length against this value before calling the
		{@link #setTray(String)} method to avoid exceptions.
		@since 3.2
	*/
	public static final int MAX_TRAY_LENGTH = BioPlateData.MAX_TRAY_LENGTH;

	/**
		The maximum length of the position coordinate that can be stored in the
		database. Check the length against this value before calling the
		{@link #setPosition(String)} method to avoid exceptions.
		@since 3.2
	*/
	public static final int MAX_POSITION_LENGTH = BioPlateData.MAX_POSITION_LENGTH;
		
	/**
		Creates a new bioplate with a specific plate geometry and type.
	  	@param dc DbControl used for database access
	  	@param geometry The geometry the new plate should have 
	  	@param plateType The type of the new plate
	  	@return A new BioPlate object
	  	@throws BaseException If anything goes wrong when creating.
	  	@throws InvalidDataException If gemoetry or type is null
	  	@since 2.16
	*/
	public static BioPlate getNew(DbControl dc, PlateGeometry geometry, BioPlateType plateType)
		throws BaseException, InvalidDataException
	{
		BioPlate bp = dc.newItem(BioPlate.class);
		bp.setPlateGeometry(geometry);
		bp.setBioPlateType(plateType);
		bp.setName("New bioplate");
		bp.getData().setEntryDate(new Date());
		return bp;
	}

	
	/**
	 	Gets a BioPlate when knowing the item's id
	  	@param dc DbControl used for database access
	  	@param id The item's id.
	  	@return <code>BioPlate</code> object
	  	@throws ItemNotFoundException If no item with the given id was found
	  	@throws PermissionDeniedException If logged in user doesn't have enough 
	  		permission to the requested item
	  	@throws BaseException If there is another error.
	 */
	public static BioPlate getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		BioPlate p = dc.loadItem(BioPlate.class, id);
		if (p == null) throw new ItemNotFoundException("BioPlate[id="+id+"]");
		return p;
	}
	
	/**
	 	Gets query that selects all available bioplates
	  	@return <code>ItemQuery</code> object
	 */
	public static ItemQuery<BioPlate> getQuery()
	{
		return new ItemQuery<BioPlate>(BioPlate.class);
	}

	BioPlate(BioPlateData annotatedData)
	{
		super(annotatedData);
	}
	
	@Override
	BioPlateData getData()
	{
		return (BioPlateData)getBasicData();
	}
	
	/**
	 	Get the freezer.
	 */
	@Override
	public Set<Annotatable> getAnnotatableParents() 
		throws BaseException
	{
		Set<Annotatable> annotatable = new HashSet<Annotatable>();
		BioPlateData data = getData();
		
		addAnnotatableParents(getDbControl(), annotatable, 
			data.getFreezer()
		);
		return annotatable;
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}	
	//-------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	/**
		@since 3.1
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------
		
	/*
	 	From the BasicItem class
	 	--------------------------------------
	 */
	/**
	 	Check if there are any <code>MeasuredBioMaterial</code> on
	 	this plate.
	 */
	@Override
	public boolean isUsed()
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIAL_ON_PLATE", "count(*)");
		/*
		 	SELECT {1}
		 	FROM MeasuredBioMaterialData mbm
		 	JOIN mbm.bioWell bw
		 	WHERE bw.bioPlate = :bioPlate
		 */
		query.setEntity("bioPlate", this.getData());
		return HibernateUtil.loadData(Long.class, query) > 0 || super.isUsed();
	}
	
	/**
	 	Gets all <code>MeasuredBioMaterial</code> located on this plate
	 	and all child plates created from this plate.
	 */
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
				"GET_BIOMATERIAL_ON_PLATE", "mbm");
		/*
		 	SELECT {1}
		 	FROM MeasuredBioMaterialData mbm
		 	JOIN mbm.BioWell bw
		 	WHERE bw.bioPlate = :bioPlate
		 */
		query.setEntity("bioPlate", this.getData());
		addUsingItems(using, query);
		return using;
	}
	@Override
	void onBeforeCommit(Action action)
		throws NotLoggedInException, BaseException
	{
		super.onBeforeCommit(action);
		if (action == Action.DELETE)
		{
			// Unlink all biomaterial events
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(
				getDbControl().getHibernateSession(), "CLEAR_PARTICIPANT_ON_BIOMATERIALEVENTS_FOR_BIOPLATE");
			/*
				UPDATE BioMaterialEventData bme
				SET bme.bioPlateEventParticipant = null
				WHERE bme.bioPlateEventParticipant =ANY (
					SELECT bep FROM BioPlateEventParticipantData bep
					WHERE bep.bioPlate = :bioPlate
				)
			*/
			query.setEntity("bioPlate", this.getData());
			query.executeUpdate();
		}
	}
	//-------------------------------------------

	/**
		Get the {@link PlateGeometry} of this bioplate.
		@return The <code>PlateGeometry</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate geometry
		@throws BaseException If there is another error
	*/
	public PlateGeometry getPlateGeometry()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateGeometry.class, getData().getPlateGeometry()); 
	}
	
	private void setPlateGeometry(PlateGeometry plateGeometry)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateGeometry == null) throw new InvalidUseOfNullException("plateGeometry");
		plateGeometry.checkPermission(Permission.USE);
		getData().setPlateGeometry(plateGeometry.getData());		
		getData().setFreeWells(plateGeometry.getColumns() * plateGeometry.getRows());
	}

	/**
		Get the {@link BioPlateType} of this bioplate.
		@return The <code>BioPlateType</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate type
		@throws BaseException If there is another error
		@since 2.16
	*/
	public BioPlateType getBioPlateType()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(BioPlateType.class, getData().getBioPlateType()); 
	}
	
	private void setBioPlateType(BioPlateType plateType)
		throws PermissionDeniedException, InvalidDataException
	{
		if (plateType == null) throw new InvalidUseOfNullException("plateType");
		plateType.checkPermission(Permission.USE);
		getData().setBioPlateType(plateType.getData());		
	}

	
	/**
	 	Sets the external id of this bio plate
	  	@param externalId The external id to set. Null is allowed.
	  		Maximum length {@link BioPlateData#MAX_EXTERNAL_ID_LENGTH} characters.
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} on the plate
	  	@throws InvalidDataException If maximum length of externalId is exceeded.
	 */
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNullableString(externalId, "externalId", BioPlateData.MAX_EXTERNAL_ID_LENGTH));
	}
	
	/**
	 	Gets the external id of this bio plate
	  	@return A String object or null
	 */
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	
	/**
	 	Sets the bio plates barcode
	  	@param barcode The barcode to set. Null is allowed.
	  		Maximum length {@link BioPlateData#MAX_BARCODE_LENGTH} characters.
	  	@throws PermissionDeniedException If logged in user doesn't have 
	  		{@link Permission#WRITE} on the plate.
	  	@throws InvalidDataException If maximum length of barcode is exceeded.
	 */
	public void setBarcode(String barcode)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setBarcode(StringUtil.setNullableString(barcode, "barcode", BioPlateData.MAX_BARCODE_LENGTH));
	}
	
	/**
	 	Gets the barcode of this bio plate.
	  	@return A String object or null.
	 */
	public String getBarcode()
	{
		return getData().getBarcode();
	}
	
	/**
		Check if the plate has been destroyed. This is currently only
		a flag. It doesn't limit what can be done with the plate through
		the API.
		@since 2.16
	*/
	public boolean isDestroyed()
	{
		return getData().isDestroyed();
	}
	
	/**
		Set the destroyed flag. 
	 	@param destroyed TRUE if the plate is destroyed, FALSE otherwise. 
	 	@throws PermissionDeniedException If the logged in user doesn't 
	 		have write permission on the item.
	 	@since 2.16
	*/
	public void setDestroyed(boolean destroyed)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setDestroyed(destroyed);
	}
	
	/**
		Get the date this plate was created.
		@return A <code>Date</code> object, or null if not known
		@since 3.1
	*/
	public Date getEventDate()
	{
		return DateUtil.copy(getData().getEventDate());
	}
	
	/**
		Set the date this plate was created.
		@param eventDate The date, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission 
		@since 3.1
	*/
	public void setEventDate(Date eventDate)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setEventDate(DateUtil.setNullableDate(eventDate, "eventDate"));
	}
	
	/**
	 	Sets the freezer where this bio plate is stored
	  	@param freezer The freezer to set. Null is allowed.
	  	@throws PermissionDeniedException If logged in user doesn't have
	  		{@link Permission#WRITE} on this bio plate or 
	  		{@link Permission#USE} on the freezer.
	 */
	public void setFreezer(Hardware freezer)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		HardwareData hw = null;
		if (freezer != null)
		{
			freezer.checkPermission(Permission.USE);
			hw = freezer.getData();
		}
		getData().setFreezer(hw);
	}
	
	/**
	 	Gets the storage location of this bio plate
	  	@return A {@link Hardware} item or null
	  	@see #getSection()
	 */
	public Hardware getFreezer()
	{
		return getDbControl().getItem(Hardware.class, getData().getFreezer());
	}
	
	/**
	 	Get the section within the "storage location" for this bioplate.
	  	@return A string or null
	  	@since 3.2
	 */
	public String getSection()
	{
		return getData().getSection();
	}
	
	/**
		Set the section within the "storage location" for this bioplate.
		@since 3.2
	*/
	public void setSection(String section)
	{
		checkPermission(Permission.WRITE);
		getData().setSection(StringUtil.setNullableString(section, "section", MAX_SECTION_LENGTH));
	}

	/**
	 	Get the tray within the "storage location" for this bioplate.
	  	@return A string or null
	  	@since 3.2
	 */
	public String getTray()
	{
		return getData().getTray();
	}
	
	/**
		Set the tray within the "storage location" for this bioplate.
		@since 3.2
	*/
	public void setTray(String tray)
	{
		checkPermission(Permission.WRITE);
		getData().setTray(StringUtil.setNullableString(tray, "tray", MAX_TRAY_LENGTH));
	}
	
	/**
	 	Get the position within the "storage location" for this bioplate.
	  	@return A string or null
	  	@since 3.2
	 */
	public String getPosition()
	{
		return getData().getPosition();
	}
	
	/**
		Set the position within the "storage location" for this bioplate.
		@since 3.2
	*/
	public void setPosition(String position)
	{
		checkPermission(Permission.WRITE);
		getData().setPosition(StringUtil.setNullableString(position, "position", MAX_POSITION_LENGTH));
	}
	
	/**
	 	Gets a query that returns all <code>BioWell</code>s 
	 	on this plate.
	  	@return an {@link ItemQuery} object.
	 */
	public ItemQuery<BioWell> getBioWells()
	{
		return BioWell.getQuery(this);
	}

	/**
		Get the number of rows on this plate. This is
		a shortcut for getPlateGeometry().getRows()
		@since 2.17
	*/
	public int getRows()
	{
		return getData().getPlateGeometry().getRows();
	}

	/**
		Get the number of rows on this plate. This is
		a shortcut for getPlateGeometry().getColumns()
		@since 2.17
	*/
	public int getColumns()
	{
		return getData().getPlateGeometry().getColumns();
	}

	/**
		Get the number of free wells on this plate.
		@since 2.17
	*/
	public int getFreeWells()
	{
		return getData().getFreeWells();
	}
	
	/**
		Get the number of used wells on this plate.
		@since 2.17
	*/
	public int getUsedWells()
	{
		return getTotalWells() - getFreeWells();
	}
	
	/**
		Get the total number of wells on this plate.
		@since 2.17
	*/
	public int getTotalWells()
	{
		PlateGeometryData pg = getData().getPlateGeometry();
		return pg.getColumns() * pg.getRows();
	}

	/**
	 	Gets bio well for specific location.
	  	@param row Row index where the bio well is located.
	  	@param column Column index where the bio well is located
	  	@return A {@link BioWell} object or 
	  		null if no bio well can be found
	 */
	public BioWell getBioWell(int row, int column)
	{
		BioWellData wellData = getData().getBioWells().get(new PlateCoordinate(row, column));
		return getDbControl().getItem(BioWell.class, wellData);
	}
	
	/**
	 	Gets bio well for specific location.
	  	@param wellC The coordinate of the well
	  	@return A {@link BioWell} object or 
	  		null if no bio well can be found
	  	@since 2.16
	 */
	public BioWell getBioWell(PlateCoordinate wellC)
	{
		if (wellC == null) return null;
		BioWellData wellData = getData().getBioWells().get(wellC);
		return getDbControl().getItem(BioWell.class, wellData);
	}
	
	/**
		Get the biomaterial in the well at the given location. This method
		is a shortcut for BioPlate.getBioWell().getBioMaterial().
	  	@param wellC The coordinate of the well
		@return A MeasuredBioMaterial or null if the well doesn't exists 
			or is empty
		@since 2.16
	 */
	public MeasuredBioMaterial getBioMaterial(PlateCoordinate wellC)
	{
		if (wellC == null) return null;
		BioWellData well = getData().getBioWells().get(wellC);
		if (well == null) return null;
		return getDbControl().getItem(MeasuredBioMaterial.class, well.getBioMaterial());
	}
	
	/**
		Get a query returning events that this plate has been a part of.
	*/
	public ItemQuery<BioPlateEvent> getEvents()
	{
		ItemQuery<BioPlateEvent> query = BioPlateEvent.getQuery();
		query.joinPermanent(Hql.innerJoin("participants", Item.BIOPLATEEVENTPARTICIPANT.getAlias()));
		query.restrictPermanent(
				Restrictions.eq(
					Hql.property(Item.BIOPLATEEVENTPARTICIPANT.getAlias(), "bioPlate"),
					Hql.entity(this)
				)
			);
		
		return query;
	}

}