/**
	$Id: DiskUsageStatistics.java 6619 2014-11-24 10:32:27Z nicklas $

	Copyright (C) 2006 Nicklas Nordborg
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Restrictions;

/**
	Get disk usage statistics for the BASE server. The statistics
	information is lazily loaded and internally cached. Ie. calling
	{@link #getSummary(User)} only hits the database the first time the 
	method is called for a user. Use the {@link #reset()} method to clear 
	the internal cache.
	<p>
	You can control how long summaries are cached by setting a maximum age
	with the {@link #setMaxAge(int)} method.

	@author nicklas
	@version 2.2
	@base.modified $Date: 2014-11-24 11:32:27 +0100 (må, 24 nov 2014) $
	@see DiskUsage#getStatistics(DbControl)
*/
public class DiskUsageStatistics
{
	private static final Long ZERO = Long.valueOf(0);

	private DbControl dc;
	private Map<Object, Summary> summaries;
	private List<QuotaType> quotaTypes;
	private int maxAge = -1;
	
	/**
		Create a new disk usage statistics object. 
	*/
	DiskUsageStatistics(DbControl dc)
	{
		this.dc = dc;
		this.summaries = new HashMap<Object, Summary>();
		
		ItemQuery<QuotaType> quotaTypeQuery = QuotaType.getQuery();
		quotaTypeQuery.order(Orders.asc(Hql.property("id")));
		quotaTypeQuery.restrict(
			Restrictions.neq(
				Hql.property("systemId"), 
				Expressions.parameter("systemId", QuotaType.TOTAL)
			)
		);
		quotaTypes = Collections.unmodifiableList(new ArrayList<QuotaType>(quotaTypeQuery.list(dc)));
	}
	
	/**
		Set the DbControl which will be used to get statisics if it is not 
		cached.
		@param dc An open DbControl object
	*/
	public void setDbControl(DbControl dc)
	{
		this.dc = dc;
	}
	
	/**
		Set the max age in seconds for a disk usage summary
		object before it is reloaded from the database again.
		A negative value keeps the object forever, a value of
		zero disabled caching.
		@param maxAge Number of seconds to keep summary in the cache,
			or zero to disable the cache or a negative value to cache
			forever
	*/
	public void setMaxAge(int maxAge)
	{
		this.maxAge = maxAge;
	}
	
	/**
		Get the max age in seconds for a disk usage summary
		object before it is reloaded from the database again.
		A negative value keeps the object forever, a value of
		zero disabled caching. The default value is -1.
	*/
	public int getMaxAge()
	{
		return maxAge;
	}
	
	/**
		Reset the internal cache of already loaded statistics.
	*/
	public void reset()
	{
		this.summaries.clear();
	}
	
	/**
		Get the list of quota types that are registered for {@link DiskConsumable}
		items.
		@return A list with QuotaType objects
	*/
	public List<QuotaType> getQuotaTypes()
	{
		return quotaTypes;
	}
	
	/**
		Get a summary of the total disk usage. This method will first check
		if a cached summary exists and only hit the database if not.
		@return A summary object
	*/
	public Summary getTotalSummary()
	{
		Summary s = summaries.get(null);
		if (s == null || s.isOld())
		{
			s = new Summary();
			if (maxAge != 0) summaries.put(null, s);
		}
		return s;
	}
	
	/**
		Get a summary of the disk usage for the specified user. This method will 
		first check if a cached summary exists and only hit the database if not.
		@param user The user to get the disk usage for
		@return A summary object
	*/
	public Summary getSummary(User user)
	{
		if (user == null) throw new InvalidUseOfNullException("user");
		Summary s = summaries.get(user);
		if (s == null || s.isOld())
		{
			s = new Summary(user);
			if (maxAge != 0) summaries.put(user, s);
		}
		return s;
	}
	
	/**
		Get a summary of the disk usage for the specified group. This method will 
		first check if a cached summary exists and only hit the database if not.
		@param group The group to get the disk usage for
		@return A summary object
	*/
	public Summary getSummary(Group group)
	{
		if (group == null) throw new InvalidUseOfNullException("group");
		Summary s = summaries.get(group);
		if (s == null || s.isOld())
		{
			s = new Summary(group);
			if (maxAge != 0) summaries.put(group, s);
		}
		return s;
	}
	
	/**
		Holds summary disk usage information for the total disk usage or for
	 	a specific user or group.
	*/
	public class Summary
	{
		private final long creationTime = System.currentTimeMillis();
		private final Map<String, Long> diskUsage;
		
		private Summary()
		{
			diskUsage = new HashMap<String, Long>();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getStatelessSession(), 
				"GET_TOTAL_DISKUSAGE_SUMMARY");
			/*
				SELECT SUM(du.bytes), du.location, du.quotaType.systemId
				FROM DiskUsageData du
				GROUP BY du.location, du.quotaType.systemId
			*/
			loadDiskUsage(query);
		}
		private Summary(User user)
		{
			diskUsage = new HashMap<String, Long>();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getStatelessSession(), 
				"GET_DISKUSAGE_SUMMARY_FOR_USER");
			/*
				SELECT SUM(du.bytes), du.location, du.quotaType.systemId
				FROM DiskUsageData du
				WHERE du.user = :user
				GROUP BY du.location, du.quotaType.systemId
			*/
			query.setInteger("user", user.getId());
			loadDiskUsage(query);
		}
		
		private Summary(Group group)
		{
			diskUsage = new HashMap<String, Long>();
			org.hibernate.Query query = HibernateUtil.getPredefinedQuery(dc.getStatelessSession(), 
				"GET_DISKUSAGE_SUMMARY_FOR_GROUP");
			/*
				SELECT SUM(du.bytes), du.location, du.quotaType.systemId
				FROM DiskUsageData du
				WHERE du.group = :group
				GROUP BY du.location, du.quotaType.systemId
			*/
			query.setInteger("group", group.getId());
			loadDiskUsage(query);		
		}
		
		private void loadDiskUsage(org.hibernate.Query query)
		{
			List<Object[]> du = HibernateUtil.loadList(Object[].class, query, null);
			for (Object[] row : du)
			{
				Long bytes = (Long)row[0];
				Location location = Location.fromValue((Integer)row[1]);
				String quotaType = (String)row[2];
				diskUsage.put(quotaType + ":" + location.name(), bytes);
				diskUsage.put(quotaType, add(diskUsage.get(quotaType), bytes));
				diskUsage.put(location.name(), add(diskUsage.get(location.name()), bytes));
				diskUsage.put(null, add(diskUsage.get(null), bytes));
			}
		}
		
		private Long add(Long a, Long b)
		{
			if (a == null) a = ZERO;
			if (b == null) b = ZERO;
			return a + b;
		}
		
		/**
			Get the age of this summary object since it was created in seconds.
			@return The age in seconds
		*/
		public int getAge()
		{
			return (int)((System.currentTimeMillis() - creationTime) / 1000);
		}
		
		/**
			Check if this summary object is old by comparing it's
			age to the {@link #getMaxAge()}.
			@return TRUE if the summary is old, FALSE otherwise
		*/
		public boolean isOld()
		{
			return maxAge > 0 && getAge() >= maxAge;
		}
		
		/**
			Get the total disk usage in bytes
		*/
		public long getTotal()
		{
			Long du = diskUsage.get(null);
			return du == null ? 0 : du.longValue();
		}

		/**
			Get the total disk usage of the specified location in bytes.
			@param location The location to get the disk usage for
		*/
		public long getTotal(Location location)
		{
			Long du = diskUsage.get(location.name());
			return du == null ? 0 : du.longValue();
		}

		/**
			Get the total disk usage for the specified quota type in bytes.
			@param quotaType The quota type to get the disk usage for
		*/
		public long getTotal(QuotaType quotaType)
		{
			Long du = diskUsage.get(quotaType.getSystemId());
			return du == null ? 0 : du.longValue();
		}

		/**
			Get the total disk usage of the specified location and
			quota type in bytes.
			@param location The location to get the disk usage for
			@param quotaType The quota type to get the disk usage for
		*/
		public long getTotal(QuotaType quotaType, Location location)
		{
			Long du = diskUsage.get(quotaType.getSystemId()+":"+location.name());
			return du == null ? 0 : du.longValue();
		}
	}
}
