/*
	$Id: News.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.NewsData;
import net.sf.basedb.core.query.EntityQuery;

import java.util.Date;

/**
	This class represents news items.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-03-13 09:15:46 +0100 (må, 13 mar 2017) $
*/
public class News
	extends BasicItem
	implements Nameable, Removable, Validatable, Registered
{
	/**
		The type of item represented by this class.
		@see Item#NEWS
		@see #getType()
	*/
	public static final Item TYPE = Item.NEWS;

	/**
		This filter limits the returned news items to those where todays date falls
		between the start and end dates unless the logged in user has generic read
		permission.
	*/
	private static final QueryRuntimeFilter RUNTIME_FILTER = new QueryRuntimeFilterImpl();
	
	/**
		Create a new <code>News</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param startDate The first day this news item should be shown
		@param newsDate The date appearing in listings of news
		@return The new <code>News</code> item
		@throws BaseException If there is an error
	*/
	public static News getNew(DbControl dc, Date startDate, Date newsDate)
		throws BaseException
	{
		News n = dc.newItem(News.class);
		n.setName("New news");
		n.setStartDate(startDate);
		n.setNewsDate(newsDate);
		n.getData().setEntryDate(new Date());
		return n;
	}

	/**
		Get a <code>News</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>News</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static News getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		News n = dc.loadItem(News.class, id);
		if (n == null) throw new ItemNotFoundException("News[id="+id+"]");
		return n;
	}

	/**
		Get a query configured to retrieve news. If the logged in user
		doesn't have generic permission to all news, or if no user is logged in,
		only news which have a start date before today and an end date after
		today are returned.
		
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<News> getQuery()
	{
		return new ItemQuery<News>(News.class, RUNTIME_FILTER);
	}

	News(NewsData newsData)
	{
		super(newsData);
	}

	@Override
	NewsData getData()
	{
		return (NewsData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Nameable interface
		-------------------------------------------
	*/
	@Override
	public String getName()
	{
		return getData().getName();
	}
	@Override
	public void setName(String name)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setName(getData(), name);
	}
	@Override
	public String getDescription()
	{
		return getData().getDescription();
	}
	@Override
	public void setDescription(String description)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		NameableUtil.setDescription(getData(), description);
	}
	// -------------------------------------------
	/*
		From the Removable interface
		-------------------------------------------
	*/
	@Override
	public boolean isRemoved()
	{
		return getData().getRemovedBy() != null;
	}
	@Override
	public void setRemoved(boolean removed)
		throws PermissionDeniedException
	{
		checkPermission(removed ? Permission.DELETE : Permission.WRITE);
		getData().setRemovedBy(removed ? getSessionControl().getLoggedInUserId() : null);
	}
	@Override
	public User getRemovedBy()
		throws PermissionDeniedException, ItemNotFoundException 
	{
		Integer userId = getData().getRemovedBy();
		return userId == null ? null : User.getById(getDbControl(), userId);
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check that startDate is before newsDate is before endDate.
	*/
	@Override
	void validate()
		throws InvalidDataException, BaseException
	{
		super.validate();
		Date startDate = getData().getStartDate();
		Date newsDate = getData().getNewsDate();
		Date endDate = getData().getEndDate();
		assert startDate != null : "startDate == null";
		assert newsDate != null : "newsDate == null";
		if (startDate.after(newsDate))
		{
			throw new InvalidDataException("Invalid date. startDate is after newsDate.");
		}
		if (endDate != null && newsDate.after(endDate)) 
		{
			throw new InvalidDataException("Invalid date. newsDate is after endDate.");
		}
	}
	/**
		Read permission is granted if today is between the start and end date.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		long today = new Date().getTime();
		long startDate = getData().getStartDate().getTime();
		long endDate = getData().getEndDate() == null ? 0 : getData().getEndDate().getTime()+24*3600*1000;
		if (startDate <= today && (endDate == 0 || today <= endDate))
		{
			granted |= Permission.grant(Permission.READ);
		}
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the start date and time for this news item to be published.
		@return The date and time
	*/
	public Date getStartDate()
	{
		return DateUtil.copy(getData().getStartDate());
	}
	/**
		Set the start date and time for this item to be published.
		@param startDate The new date and time
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission
		@throws InvalidDataException If the date is null
	*/
	public void setStartDate(Date startDate)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setStartDate(DateUtil.setNotNullDate(startDate, "startDate"));
	}

	/**
		Get the date that is shown in listings.
		@return The date and time
	*/
	public Date getNewsDate()
	{
		return DateUtil.copy(getData().getNewsDate());
	}
	/**
		Set the date that is shown in listings.
		@param newsDate The new date and time
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission
		@throws InvalidDataException If the date is null
	*/
	public void setNewsDate(Date newsDate)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setNewsDate(DateUtil.setNotNullDate(newsDate, "newsDate"));
	}

	/**
		Get the end date and time for this news item to be published. If this value
		is null, the new item is published forever.
		@return The date and time
	*/
	public Date getEndDate()
	{
		return DateUtil.copy(getData().getEndDate());
	}
	/**
		Set the end date and time for this item to be published.
		@param endDate The new date and time, or null to publish this new item forever
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#WRITE} permission
	*/
	public void setEndDate(Date endDate)
		throws PermissionDeniedException
	{
		checkPermission(Permission.WRITE);
		getData().setEndDate(DateUtil.setNullableDate(endDate, "endDate"));
	}

	private static class QueryRuntimeFilterImpl
		implements QueryRuntimeFilter
	{
		@Override
		public void enableFilters(QueryRuntimeFilterManager manager, EntityQuery query, DbControl dc)
		{
			SessionControl sc = dc.getSessionControl();
			if (sc.isDenied(query.getItemPermission(), Item.NEWS))
			{
				manager.enableFilter("denyAll");
			}
			else if (!sc.hasPermission(query.getItemPermission(), Item.NEWS))
			{
				if (query.getItemPermission() == Permission.READ)
				{
					org.hibernate.Filter filter = manager.enableFilter("todaysNews");
					if (filter != null)
					{
						filter.setParameter("today", new Date());
					}
				}
				else
				{
					manager.enableFilter("denyAll");
				}
			}
		}
	}
}

