/*
	$Id: PlateMapping.java 7308 2017-03-13 08:15:46Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.PlateMappingData;
import net.sf.basedb.core.data.PlateGeometryData;
import net.sf.basedb.core.data.PlateData;
import net.sf.basedb.core.data.WellData;
import net.sf.basedb.core.data.MappingCoordinate;
import net.sf.basedb.core.data.PlateCoordinate;

import java.util.Date;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;

/**
	This class represents a mapping from one plate geometry to another.
	Given one or more plates of some specific geometry,
	a mapping creates one or more new plates of some other or the same
	geometry. The new plates and their wells will be linked to the
	plates and wells they were created from.
	<p>
	A mapping is specified by a set of source/destination coordinate
	pairs. A coordinate is specified by plate number, row and column.
	With this information it is possible to pick wells from the
	source plates and link them to wells on the destination plates.
	<p>
	It is not required that the mapping is complete, ie. contains
	a mapping for each destination coordinate. Such destination wells
	will be marked as empty. A destination coordinate may only be mapped to
	one source coordinate, but it is possible that one source coordinate
	is mapped to several destination coordinates.

	@author Nicklas
	@version 2.0
	@see Plate
	@see PlateGeometry
	@base.modified $Date: 2017-03-13 09:15:46 +0100 (må, 13 mar 2017) $
*/
public class PlateMapping
	extends CommonItem
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#PLATEMAPPING
		@see #getType()
	*/
	public static final Item TYPE = Item.PLATEMAPPING;
	
	/**
		Create a new <code>PlateMapping</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param sourceGeometry The required geometry of the source plates
		@param sourceCount The number of source plates
		@param destinationGeometry The geometry of the created destination plates
		@param destinationCount The number of created destination plates
		@return The new <code>PlateMapping</code> item
		@throws PermissionDeniedException If the loggged in user
			doesn't have use permission for the geometries
		@throws InvalidDataException If any of the geometries are null, or
			any of the count values are less than one
		@throws BaseException If there is another error
	*/
	public static PlateMapping getNew(DbControl dc, 
		PlateGeometry sourceGeometry, int sourceCount, PlateGeometry destinationGeometry, int destinationCount)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		PlateMapping pm = dc.newItem(PlateMapping.class);
		pm.setName("New plate mapping");
		pm.setSourceGeometry(sourceGeometry);
		pm.setSourceCount(sourceCount);
		pm.setDestinationGeometry(destinationGeometry);
		pm.setDestinationCount(destinationCount);
		pm.getData().setEntryDate(new Date());
		return pm;
	}

	/**
		Get a <code>PlateMapping</code> item when you know the id.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The id of the item to load
		@return The <code>PlateMapping</code> item
		@throws ItemNotFoundException If an item with the specified 
			id is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static PlateMapping getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		PlateMapping pm = dc.loadItem(PlateMapping.class, id);
		if (pm == null) throw new ItemNotFoundException("PlateMapping[id="+id+"]");
		return pm;
	}

	/**
		Get a query that returns plate mappings.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<PlateMapping> getQuery()
	{
		return new ItemQuery<PlateMapping>(PlateMapping.class);
	}

	PlateMapping(PlateMappingData plateMappingData)
	{
		super(plateMappingData);
	}
	
	@Override
	PlateMappingData getData()
	{
		return (PlateMappingData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		Check if:
		<ul>
		<li>A {@link Plate} has been created from this mapping
		</ul>
	*/
	@Override
	public boolean isUsed()
		throws BaseException
	{
		org.hibernate.Session session = getDbControl().getHibernateSession();
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATES_FOR_PLATEMAPPING", "count(*)");
		/*
			SELECT {1}
			FROM PlateData p
			WHERE p.plateMapping = :plateMapping
		*/
		query.setEntity("plateMapping", this.getData());
		boolean used = HibernateUtil.loadData(Long.class, query) > 0;
		return used || super.isUsed();
	}
	/**
		Get all:
		<ul>
		<li>{@link Plate}:s created with this mapping
		</ul>
		@since 2.2
	*/
	@Override
	public Set<ItemProxy> getUsingItems()
	{
		Set<ItemProxy> using = super.getUsingItems();
		org.hibernate.Session session = getDbControl().getHibernateSession();
		// Plates
		org.hibernate.Query query = HibernateUtil.getPredefinedQuery(session, 
			"GET_PLATES_FOR_PLATEMAPPING", "p.id");
			/*
				SELECT {1}
				FROM PlateData p
				WHERE p.plateMapping = :plateMapping
			*/
		query.setEntity("plateMapping", this.getData());
		addUsingItems(using, Item.PLATE, query);
		return using;
	}
	// -------------------------------------------

	/**
		Get the {@link PlateGeometry} that is required for source plates.
		@return A <code>PlateGeometry</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate geometry
		@throws BaseException If there is another error
	*/
	public PlateGeometry getSourceGeometry()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateGeometry.class, getData().getSourceGeometry()); 
	}

	/**
		Set the source geometry.
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for the geometry
		@throws InvalidDataException If the geometry is null
	*/
	private void setSourceGeometry(PlateGeometry sourceGeometry)
		throws PermissionDeniedException, InvalidDataException
	{
		if (sourceGeometry == null) throw new InvalidUseOfNullException("sourceGeometry");
		sourceGeometry.checkPermission(Permission.USE);
		getData().setSourceGeometry(sourceGeometry.getData());
	}
	
	/**
		Get the number of required source plates.
	*/
	public int getSourceCount()
	{
		return getData().getSourceCount();
	}
	
	/**
		Set the number of source plates.
		@throws InvalidDataException If the source count is less than one
	*/
	private void setSourceCount(int sourceCount)
		throws InvalidDataException
	{
		getData().setSourceCount(IntegerUtil.checkMin(sourceCount, "sourceCount", 1));
	}

	/**
		Get the {@link PlateGeometry} of the created plates.
		@return A <code>PlateGeometry</code> item
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ} permission for the plate geometry
		@throws BaseException If there is another error
	*/
	public PlateGeometry getDestinationGeometry()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(PlateGeometry.class, getData().getDestinationGeometry()); 
	}
	
	/**
		Set the destination geometry.
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for the geometry
		@throws InvalidDataException If the geometry is null
	*/
	private void setDestinationGeometry(PlateGeometry destinationGeometry)
		throws PermissionDeniedException, InvalidDataException
	{
		if (destinationGeometry == null) throw new InvalidUseOfNullException("destinationGeometry");
		destinationGeometry.checkPermission(Permission.USE);
		getData().setDestinationGeometry(destinationGeometry.getData());
	}

	/**
		Get the number of produced destination plates.
	*/
	public int getDestinationCount()
	{
		return getData().getDestinationCount();
	}
	
	/**
		Set the number of produced destination plates.
		@throws InvalidDataException If the destination count is less than one
	*/
	private void setDestinationCount(int destinationCount)
		throws InvalidDataException
	{
		getData().setDestinationCount(IntegerUtil.checkMin(destinationCount, "destinationCount", 1));
	}

	/**
		The maximum length of the image name that can be stored in the
		database. Check the length against this value before calling the
		{@link #setImage(String)} method to avoid exceptions.
	*/
	public static final int MAX_IMAGE_LENGTH = PlateMappingData.MAX_IMAGE_LENGTH;
	/**
		Get the name of a small image that is a graphical representation of the mapping.
		@return The name of an image or null
	*/
	public String getImage()
	{
		return getData().getImage();
	}
	/**
		Set the name of a small image that is a graphical representation of the mapping.
		Null is allowed but the value can't be longer than the value specified by
		the {@link #MAX_IMAGE_LENGTH} constant. 
		
		@param image The name of the image
		@throws PermissionDeniedException If the logged in user doesn't
			have write permission
		@throws InvalidDataException If the value is too long
	*/
	public void setImage(String image)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setImage(StringUtil.setNullableString(image, "image", MAX_IMAGE_LENGTH));
	}

	/**
		Get the source coordinate for a given destination coordinate.
		@param destinationCoordinate The destination coordinate
		@return A <code>MappingCoordinate</code> object or null if
			no source coordinate has been specified
	*/
	public MappingCoordinate getSourceCoordinate(MappingCoordinate destinationCoordinate)
	{
		return getData().getDetails().get(destinationCoordinate);
	}
	
	/**
		Set the source coordinate for a given destination coordinate.
		@param destinationCoordinate The destination coordinate
		@param sourceCoordinate The source coordinate, or null to remove the mapping
			for the destination coordinate
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission or if the mapping has been saved to the database
		@throws InvalidDataException If the destination coordinate is null or
			any of the coordinates are outside the range of valid coordinates
			as specified by the geometries
	*/
	public void setSourceCoordinate(MappingCoordinate destinationCoordinate, MappingCoordinate sourceCoordinate)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		if (isInDatabase())
		{
			throw new PermissionDeniedException("Not allowed to modify the coordinates for an existing mapping");
		}
		if (destinationCoordinate == null) throw new InvalidUseOfNullException("destinationCoordinate");
		checkCoordinate(destinationCoordinate, getDestinationCount(), getData().getDestinationGeometry(), true);
		if (sourceCoordinate == null)
		{
			getData().getDetails().remove(destinationCoordinate);
		}
		else
		{
			checkCoordinate(sourceCoordinate, getSourceCount(), getData().getSourceGeometry(), false);
			getData().getDetails().put(destinationCoordinate, sourceCoordinate);
		}
	}

	/**
		Checks that the coordinate is within the bouds given by the mapping
		and geometry.
	*/
	private void checkCoordinate(MappingCoordinate c, int plateCount, PlateGeometryData geometry, boolean destination)
		throws InvalidDataException
	{
		if (c.getPlate() < 0 || c.getPlate() >= plateCount
			|| c.getRow() < 0 || c.getRow() >= geometry.getRows()
			|| c.getColumn() < 0 || c.getColumn() >= geometry.getColumns())
		{
			throw new InvalidDataException("Invalid "+(destination ? "destination" : "source")+ "coordinate: " + c);
		}
	}
	
	
	/**
		Apply this mapping to the given list of source plates and generate
		new plates of the specified type. The source plates must all be
		of the geometry specified by {@link #getSourceGeometry()} and the destroyed
		flag mustn't be set. The list must contain exactly the required number 
		of plates as specified by @link #getSourceCount()}.
		The destination plate type must be of the required destination geometry, 
		{@link #getDestinationGeometry()}.
		<p>
		The new plates are not saved to the database and doesn't have any
		names, descriptions or other information. It is up to the client
		application to fill in this information and save the new plates.

		@param sourcePlates A <code>List</code> containing {@link Plate}
			items
		@param destinationPlateType The {@link PlateType} of the
			destination plates
		@return A <code>List</code> containing the new (unsaved) {@link Plate}
			items
		@throws PermissionDeniedException If the logged in user doesn't have
			use permission for this plate mapping and to all source plates
		@throws InvalidDataException If any of the parameters are invalid
		@throws BaseException If there is another error
	*/
	public List<Plate> apply(List<Plate> sourcePlates, PlateType destinationPlateType)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.USE);
		// Check that the number of source plates are correct
		if ((sourcePlates == null) || (sourcePlates.size() != getSourceCount()))
		{
			throw new InvalidDataException("Invalid number of source plates."+
				" Expected: "+getSourceCount()+
				" Found: "+(sourcePlates == null ? 0 : sourcePlates.size())
			);
		}
		// Check that all source plates have the correct geometry
		// and that none of them are flagged as destroyed, null is allowed if not 
		// all source plate positions are used
		PlateGeometryData sourceGeometry = getData().getSourceGeometry();
		Map<PlateData, Integer> sourcePlatesMap = new HashMap<PlateData, Integer>(sourcePlates.size());
		for (int sourceIndex = 0; sourceIndex < sourcePlates.size(); sourceIndex++)
		{
			Plate source = sourcePlates.get(sourceIndex);
			if (source != null)
			{
				source.checkPermission(Permission.USE);
				if (!sourceGeometry.equals(source.getData().getPlateType().getPlateGeometry()))
				{
					throw new InvalidDataException("Invalid plate geometry of source: "+source);
				}
				if (source.isDestroyed())
				{
					throw new InvalidDataException("Plate is flagged as destroyed: "+source);
				}
				sourcePlatesMap.put(source.getData(), sourceIndex);
			}
		}
		// Check that the destination plate type has the correct geometry
		if (destinationPlateType == null) throw new InvalidUseOfNullException("destinationPlateType");
		PlateGeometryData destinationGeometry = getData().getDestinationGeometry();
		if (!destinationGeometry.equals(destinationPlateType.getData().getPlateGeometry()))
		{
			throw new InvalidDataException("Invalid plate geometry of destination type: "+destinationPlateType);
		}

		// Create the new plates
		List<Plate> destinationPlates = new ArrayList<Plate>(getDestinationCount());
		DbControl dc = getDbControl();
		for (int j = 0; j < getDestinationCount(); ++j)
		{
			Plate p = Plate.getNew(dc, destinationPlateType);
			PlateData pData = p.getData();
			pData.setPlateMapping(getData());
			pData.setDestinationIndex(j);
			pData.getParents().putAll(sourcePlatesMap);
			destinationPlates.add(p);
		}

		// Link the wells of the destination plates to the source plates
		for (Map.Entry<MappingCoordinate, MappingCoordinate> entry : getData().getDetails().entrySet())
		{
			MappingCoordinate destC = entry.getKey();
			MappingCoordinate sourceC = entry.getValue();

			Plate destination = destinationPlates.get(destC.getPlate());
			Plate source = sourcePlates.get(sourceC.getPlate());

			if (source != null)
			{
				WellData destWell = destination.getData().getWells().get(new PlateCoordinate(destC.getRow(), destC.getColumn()));
				WellData sourceWell = source.getData().getWells().get(new PlateCoordinate(sourceC.getRow(), sourceC.getColumn()));
				destWell.setParent(sourceWell);
				destWell.setReporter(sourceWell.getReporter());
			}
		}

		return destinationPlates;
	}

}
