/*
	$Id: PluginConfigurationRequest.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson, Gregory Vincic

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.plugin.ImmediateDownloadExporter;
import net.sf.basedb.core.plugin.Plugin;
import net.sf.basedb.core.plugin.InteractivePlugin;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;

import java.util.List;

/**
	Configures a plugin or job. You will get an instance of
	this class when:
	
	<ol>
	<li>You want to configure a plugin: {@link PluginConfiguration#configure()}
	<li>You want to configure a job: {@link Job#configure(GuiContext)}
	</ol>
	
	You should use the following protocol:
	<ol>
	<li>Call the {@link #getRequestInformation()} method
		to find out which parameters the plugin or job wants.
	
	<li>Display a user interface for entering the parameter values or get the
		parameters from somewhere else.
	
	<li>Set the parameter values with the {@link #setParameterValue(String, Object)}
		or {@link #setParameterValues(String, List)} methods.

	<li>Call {@link #invoke()} to let the plugin check the parameters and
		store them in the database.

	<li>Check the {@link PluginResponse#getStatus()} method and repeat the 
		steps if needed, or display an error message. In case of an error
		in the parameters, ask for new parameters and retry.

	<li>Commit the changes using {@link DbControl#commit()}.
	</ol>
	
	Note 1! If the above procedure includes waiting for user input (which it normally does)
	you should also use {@link DbControl#disconnect()} and {@link DbControl#reconnect()}
	to conserve database resources. You must not close or commit the <code>DbControl</code>
	until the configuration is completed.
	
	<p>
	Note 2! If the returned status of a plugin is {@link net.sf.basedb.core.plugin.Response.Status#DONE} after
	an invokation, the {@link Plugin#done()} method is called to let the plugin 
	clean up resources. If a client application wants to abort earlier it should
	use the {@link #done()} method in this class.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2015-04-21 09:59:42 +0200 (ti, 21 apr 2015) $
*/
public final class PluginConfigurationRequest
	extends PluginRequest<InteractivePlugin>
{

	private GuiContext context;
	private Job.ProgressReporterImpl progress;
	private boolean allowImmediateExecution;
	private boolean dryRun;
	
	PluginConfigurationRequest(SessionControl sc, InteractivePlugin plugin, String command,
		PluginConfiguration pluginConfiguration, ParameterValuesImpl configurationParameters,
		Job job, ParameterValuesImpl jobParameters, GuiContext context, 
		PluginDefinition pluginDefinition, boolean allowImmediateExecution)
	{
		super(sc, plugin, command, pluginConfiguration, configurationParameters, job, jobParameters, pluginDefinition);
		this.context = context;
		this.allowImmediateExecution = allowImmediateExecution;
		if (job != null) dryRun = job.isDryRun();
	}
	
	/*
		From the PluginRequest class
		-------------------------------------------
	*/
	@Override
	public PluginResponse invoke()
	{
		InteractivePlugin plugin = getPlugin();
		boolean allowImmediateDownload = allowImmediateExecution && 
			plugin instanceof ImmediateDownloadExporter;
		PluginResponse pluginResponse = new PluginResponse(this, allowImmediateExecution, allowImmediateDownload, dryRun);
		Request request = new RequestImpl(allowImmediateExecution);
		Response response = pluginResponse.getResponseImpl();
		try
		{
			plugin.configure(getGuiContext(), request, response);
		}
		catch (Throwable t)
		{
			pluginResponse.setError("Error invoking plugin", t);
		}
		if (pluginResponse.getStatus().allowSaveParameters())
		{
			if (pluginResponse.getStatus() == Response.Status.DONE) done();
			if (progress != null) progress.setConfigured(pluginResponse);
		}
		return pluginResponse;
	}
	// -------------------------------------------

	/**
		Get the request information object containing 
		information about the parameters needed for the plugin.
		@return A <code>RequestInformation</code> object
		@throws BaseException If there is an error
	*/
	public RequestInformation getRequestInformation()
		throws BaseException
	{
		RequestInformation ri = getPlugin().getRequestInformation(getGuiContext(), getCommand());
		registerParameters(ri.getParameters());
		return ri;
	}
	
	/**
		Get the context of the client application when this configuration request
		was created.
	*/
	public GuiContext getGuiContext()
	{
		return context;
	}
	
	void setProgressReporter(Job.ProgressReporterImpl progress)
	{
		this.progress = progress;
	}
}
