/*
	$Id: Session.java 7165 2016-06-03 11:30:31Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.SessionData;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Hql;

import java.util.Date;

/**
	A session represents the time between login and logout for a single
	user. Hence, a session item is automatically created when a user
	logs in, and is automatically updated when the user logs out.
	<p>
	A session said to be active when the user is logged in. 

	@author Nicklas
	@version 2.0
*/
public class Session
	extends BasicItem
{
	/**
		The type of item represented by this class.
		@see Item#SESSION
		@see #getType()
	*/
	public static final Item TYPE = Item.SESSION;

	/**
		Get a <code>Session</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Session</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Session getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Session s = dc.loadItem(Session.class, id);
		if (s == null) throw new ItemNotFoundException("Session[id="+id+"]");
		return s;
	}

	/**
		Get a query configured to retrieve sessions for the specified user.
		
		@param user The user to retreive sessions for, null is allowed if
			the logged in user has generic READ permission for sessions in which case
			all sessions will be returned
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Session> getQuery(User user)
	{
		ItemQuery<Session> query = null;
		
		if (user != null)
		{
			query = new ItemQuery<Session>(Session.class, null);
			query.restrictPermanent(
				Restrictions.eq(
					Hql.property("user"), 
					Hql.entity(user)
				)
			);
		}
		else
		{
			query = new ItemQuery<Session>(Session.class);
		}
		return query;
	}

	Session(SessionData sessionData)
	{
		super(sessionData);
	}

	@Override
	SessionData getData()
	{
		return (SessionData)getBasicData();
	}
	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------

	/*
		From the BasicItem class
		-------------------------------------------
	*/
	/**
		READ permission is granted if the logged in user is the owner of this
		session. All other permissions are denied.
	*/
	@Override
	void initPermissions(int granted, int denied)
		throws BaseException
	{
		if (getSessionControl().getLoggedInUserId() == getData().getUser().getId())
		{
			granted |= Permission.grant(Permission.READ);
		}
		denied |= Permission.deny(Permission.WRITE);
		super.initPermissions(granted, denied);
	}
	// -------------------------------------------

	/**
		Get the user that logged in with this session.

		@return The <code>User</code> item
		@throws PermissionDeniedException If the logged in user doesn't have 
			{@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public User getUser()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(User.class, getData().getUser());
	}

	/**
		Get the client application the user was using in this session.

		@return A <code>Client</code> item, or null if not known
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission
		@throws BaseException If there is another error
	*/
	public Client getClient()
		throws PermissionDeniedException, BaseException
	{
		return getDbControl().getItem(Client.class, getData().getClient());
	}

	/**
		Get the date and time the user logged in.
		@return A <code>Date</code> object with the login time
	*/
	public Date getLoginTime()
	{
		return DateUtil.copy(getData().getLoginTime());
	}

	/**
		Get the date and time the user logged out.
		@return A <code>Date</code> object with the logout time
	*/
	public Date getLogoutTime()
	{
		return DateUtil.copy(getData().getLogoutTime());
	}

	/**
		Get the login comment.
		@return A <code>String</code> object with the login comment
	*/
	public String getLoginComment()
	{
		return getData().getLoginComment();
	}

	/**
		Check if another user was acting as the user of this session.
		@return TRUE if the user was impersonated, FALSE otherwise
	*/
	public boolean getImpersonated()
	{
		return getData().getImpersonated();
	}

	/**
		Get the remote ID of the host the user used for this session.
		Typically it is the IP-address of the user's computer.
		@return A <code>String</code> object with remote ID
	*/
	public String getRemoteId()
	{
		return getData().getRemoteId();
	}

	/**
		Get the authentication method that was used to login to
		this session.
		@since 3.9
	*/
	public String getAuthenticationMethod()
	{
		return getData().getAuthenticationMethod();
	}
}



