/**
	$Id: MySQLEngine.java 7075 2016-02-15 10:58:10Z nicklas $

	Copyright (C) 2006, 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	BASE is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.dbengine;

import java.util.Collection;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.hibernate.mapping.PersistentClass;

/**
	Database engine for My SQL.

	@author nicklas
	@version 2.0
	@base.modified $Date: 2016-02-15 11:58:10 +0100 (må, 15 feb 2016) $
*/
public class MySQLEngine
	extends AbstractDbEngine
	implements DbEngine
{
	/**
		Create MySQLEngine.
	*/
	public MySQLEngine()
	{}
	
	/*
		From the DbEngine interface
		-------------------------------------------
	*/
	/**
		Always false.
	*/
	@Override
	public boolean useThetaJoin()
	{
		return false;
	}
	
	/**
		Generates a catalog name by combining the catalog and
		schema name (catalog_schema).
		@since 3.4
	*/
	@Override
	public String getCatalogName(String catalog, String schema)
	{
		return catalog + "_" + schema;
	}
	
	/**
		Return null.
		@since 3.4
	*/
	@Override
	public String getSchemaName(String catalog, String schema)
	{
		return null;
	}
	
	/**
		Generate <code>create schema &lt;catalog&gt;</code>.
		@since 3.4
	*/
	@Override
	public String getCreateSchemaSql(String catalog, String schema)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("create schema ").append(getQuotedName(catalog));
		return sql.toString();
	}

	
	/**
	 	Generate <code>alter table &lt;catalog&gt;.&lt;table&gt; add [unique] 
	 	index &lt;name&gt; (&lt;columns&gt;)</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getCreateIndexSql(String catalog, String schema, String table, String name, 
		Set<String> columns, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("alter table ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table)).append(" add ");
		if (unique) sql.append("unique ");
		sql.append("index ").append(getQuotedName(name)).append(" (");
		appendList(sql, columns);
		sql.append(")");
		return sql.toString();
	}
	
	/**
		@return TRUE
		@since 3.6
	*/
	@Override
	public boolean dropForeignKeysUsedInIndex()
	{
		return true;
	}
	
	/**
	 	Generate <code>drop index &lt;name&gt; on &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getDropIndexSql(String catalog, String schema, String table, String name, boolean unique)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("drop index ").append(getQuotedName(name)).append(" on ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		return sql.toString();
	}

	/**
		Generate <code>alter table schema.table add constraint name
		foreign key (columns) references refTable (refColumns)</code>
		@since 3.1
	 */
	@Override
	public String getCreateForeignKeySql(String catalog, String schema,	String table, 
		String name, Set<String> columns, String refTable, Set<String> refColumns) 
	{
		StringBuilder sql = new StringBuilder();
		
		sql.append("alter table ");
		if (schema != null) sql.append(getQuotedName(schema)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" add constraint ");
		sql.append(getQuotedName(name));
		sql.append(" foreign key (");
		appendList(sql, columns);
		sql.append(") references ");
		sql.append(getQuotedName(refTable));
		sql.append(" (");
		appendList(sql, refColumns);
		sql.append(")");
		return sql.toString();
	}
	
	/**
		Generate <code>alter table &lt;catalog&gt;.&lt;table&gt; drop foreign key &lt;name&gt;</code>
		@since 3.0
	*/
	@Override
	public String getDropForeignKeySql(String catalog, String schema, String table, String name)
	{
		StringBuilder sql = new StringBuilder();
		sql.append("alter table ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" drop foreign key ");
		sql.append(getQuotedName(name));
		return sql.toString();
	}
	
	/**
		Generate <code>alter table catalog.table add primary key (columns)</code>
		@since 3.1
	*/
	@Override
	public String getCreatePrimaryKeySql(String catalog, String schema, String table, String name, Set<String> columns) 
	{
		StringBuilder sql = new StringBuilder();
		sql.append("alter table ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" add primary key (");
		appendList(sql, columns);
		sql.append(")");
		return sql.toString();
	}

	/**
		Generate <code>alter table catalog.table drop primary key</code>
		@since 3.1
	*/
	@Override
	public String getDropPrimaryKey(String catalog, String schema, String table, String name) 
	{
		StringBuilder sql = new StringBuilder();
		sql.append("alter table ");
		if (catalog != null) sql.append(getQuotedName(catalog)).append(".");
		sql.append(getQuotedName(table));
		sql.append(" drop primary key");
		return sql.toString();
	}

	/**
		@return Always "null" (assuming that the ID column is an AUTO_INCREMENT column)
	*/
	@Override
	public String getInsertAutoIncrementSql(PersistentClass pc)
	{
		return "null";
	}

	/**
		Return true.
	*/
	@Override
	public boolean analyzeAfterBatchOperation()
	{
		return true;
	}

	/**
		Generate <code>optimize table &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getOptimizeTableSql(String catalog, String schema, String table)
	{
		catalog = catalog == null ? "" : getQuotedName(catalog) + ".";
		return "optimize table " + catalog + getQuotedName(table);
	}
	
	/**
		Generate <code>analyze table &lt;catalog&gt;.&lt;table&gt;</code>.
		The schema parameter is not supported.
	*/
	@Override
	public String getAnalyzeTableSql(String catalog, String schema, String table)
	{
		catalog = catalog == null ? "" : getQuotedName(catalog) + ".";
		return "analyze table " + catalog + getQuotedName(table);
	}

	/**
		Replace "create table" with "create table if not exists"...
		@since 2.6
	*/
	@Override
	public String makeSafeCreateTable(String sql, String catalog, String schema, String table)
	{
		Pattern p = Pattern.compile("create table", Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(sql);
		sql = m.replaceFirst("create table if not exists");
		return sql;
	}
	
	/**
		Returns TRUE.
		@since 2.4
	*/
	@Override
	public boolean caseInsensitiveComparison()
	{
		return true;
	}
		
	
	/**
		Returns FALSE.
	*/
	@Override
	public boolean supportsColumnAliasInWhere()
	{
		return false;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInOrderBy()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInGroupBy()
	{
		return true;
	}
	/**
		Returns TRUE.
	*/
	@Override
	public boolean supportColumnAliasInHaving()
	{
		return true;
	}
	/**
		Return <code>LOG(&lt;value&gt;)</code>.
	*/
	@Override
	public String ln(String value)
	{
		return "LOG("+ value + ")";
	}
	/**
		Return <code>&lt;value&gt; RLIKE &lt;regexp&gt;</code>.
		@since 2.8
	*/
	@Override
	public String rlike(String value, String regexp)
	{
		return value + " RLIKE " + regexp;
	}
	
	/**
		Return <code>varchar(length) CHARACTER SET utf8 COLLATE utf8_bin</code>.
		@since 2.9
	 */
	@Override
	public String getCaseSensitiveVarchar(int length)
	{
		return "varchar(" + length + ") CHARACTER SET utf8 COLLATE utf8_bin";
	}
	/**
		Put quotes (`) around the name.
	*/
	@Override
	public String getQuotedName(String name)
	{
		return "`" + name + "`";
	}
	
	@Override
	public String getApproximateRowCountSql(String catalog, String schema, String table)
	{
		return "select TABLE_ROWS from INFORMATION_SCHEMA.TABLES where TABLE_NAME='" + table + "' and TABLE_SCHEMA='" + catalog + "'";
	}

	// -------------------------------------------

	private void appendList(StringBuilder sb, Collection<String> items)
	{
		int i = 0;
		for (String column : items)
		{
			if (i > 0) sb.append(", ");
			sb.append(getQuotedName(column));
			i++;
		}
	}

}
