/*
	$Id: Diff3.java 4889 2009-04-06 12:52:39Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.ItemModifiedException;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;

/**
	This class is helpful when a client application needs to implement long-running
	transactions and still be able to update items with as few clashes as possible.
	Using this class it is possible to avoid {@link ItemModifiedException}:s
	even when two or more users are modifying the same item at the same time. 
	The only requirement is that they don't change the same property to different 
	values. Here is how to use this class:
	<ul>
	<li>The <code>oldItem</code> is loaded and the user is presented with a
		screen to modify the properties of the item. The client application must 
		keep the item in memory while the user is doing the modifications.
	<li>When the user submits the changes the <code>newItem</code> is loaded.
		Both objects are passed to the {@link #Diff3(BasicItem, BasicItem)} constructor.
		The construcor checks if both objects are of the same version or not.
	<li>The client application uses the {@link #getValue(Object, Object, Object)} method to
		set the properties on the <code>newItem</code> object.
	</ul>
	
	<pre class="code">
<b>edit.jsp</b>
Label label = Label.getById(dbControl, labelId);
sessionControl.setSessionSetting("LABEL", label);

<b>submit.jsp</b>
Label oldLabel = (Label)sessionControl.getSessionSetting("LABEL");
Label newLabel = Label.getById(dbControl, labelId);
Diff3 d3 = new Diff3(oldLabel, newLabel);
newLabel.setName(d3.getValue(oldLabel.getName(), newLabel.getName(), 
   request.getParameter("name")));
// ... etc ...
dbControl.commit();
</pre>

	Additional code to check for session timeouts and creating new items
	is also required.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2009-04-06 14:52:39 +0200 (må, 06 apr 2009) $
*/
public class Diff3
{
	private final boolean sameVersion;
	private final String what;

	/**
		Create a new <code>Diff3</code> object to compare properties
		of the <code>oldItem</code> and <code>newItem</code>. Both
		objects mustn't be null and they must represent the same
		item in the database. For new objects, not yet in the database,
		both parameters must be the same instance.
		@param oldItem The "old" representation of the item
		@param newItem The "new" representation of the item
		@throws InvalidDataException If one of the parameters is null
			or they don't represent the same database item
	*/
	public Diff3(BasicItem oldItem, BasicItem newItem)
		throws InvalidDataException
	{
		if (oldItem == null) throw new InvalidUseOfNullException("oldItem");
		if (newItem == null) throw new InvalidUseOfNullException("newItem");
		if (!oldItem.equals(newItem)) 
		{
			throw new InvalidDataException("oldItem '"+oldItem+"' is not the same item as newItem '"+newItem+"'");
		}
		sameVersion = oldItem.getVersion() == newItem.getVersion();
		what = newItem.toString();
	}
	
	/**
		Get the correct value to set on the <code>newItem</code> object.
		<ul>
		<li>If the old and new item have the same version 
			or if the <code>oldValue</code> is the same as the
			<code>newValue</code> the <code>requestValue</code> 
			is returned. This represents the case that no other
			user has modified the item, or it was modified but not
			this particular property.
		<li>If the old and request value are equal the <code>newValue</code>
			is returned. This represents the case that the other
			user has modified the property, but the current user hasn't.
		<li>If the new and request value are equal the <code>newValue</code>
			is returned. This represents the case that both users
			has modified the property and, fortunately, set it to the
			same value.
		<li>Otherwise, it all three values are different, a
			{@link ItemModifiedException} is thrown.
		</ul>
		@param oldValue The value of the property on the old item
		@param newValue The value of the property on the new item
		@param requestValue The vale of the property the current user 
			has set on the edit screen
		@return The most appropriate value as described above
		@throws ItemModifiedException If all values are different
	*/
	public <T> T getValue(T oldValue, T newValue, T requestValue)
		throws ItemModifiedException
	{
		if (sameVersion || isEqualOrNull(oldValue, newValue))
		{
			return requestValue;
		}
		else if (isEqualOrNull(oldValue, requestValue))
		{
			return newValue;
		}
		else if (isEqualOrNull(newValue, requestValue))
		{
			return newValue;
		}
		throw new ItemModifiedException(what);
	}
	
	/**
		Check if two objects are both null or equal as defined
		by the {@link Object#equals(Object)} method.
	 	@param o1 Object to compare 
	 	@param o2 Object to compare
	 	@return TRUE if equal or null, FALSE otherwise.
	*/
	public static boolean isEqualOrNull(Object o1, Object o2)
	{
		return (o1 == o2) || (o1 != null && o1.equals(o2));
	}
	
}
