/**
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.annotations;

import net.sf.basedb.core.Annotatable;
import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateType;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.snapshot.AnnotationSnapshot;
import net.sf.basedb.util.filter.Filter;

/**
	Class for holding rules for how annotations should be inherited from
	parent items. If {@link #getRemove()} is set, it override all other
	settings and all inherited annotation for the specified annotation type
	are removed.
	
	@author nicklas
	@since 3.5
*/
public class InheritSpecification 
	implements Filter<AnnotationSnapshot>
{

	private final int annotationTypeId;
	private final AnnotationType annotationType;

	private boolean remove;
	private boolean clone;
	private boolean resync;
	private ItemSubtype subtype;
	private boolean noDuplicates = true;
	private boolean replaceExisting = false;
	
	/**
		Creates a new specification for the given annotation type.
	*/
	public InheritSpecification(AnnotationType annotationType)
	{
		this.annotationTypeId = annotationType.getId();
		this.annotationType = annotationType;
	}
	
	public AnnotationType getAnnotationType()
	{
		return annotationType;
	}
	
	/**
		Set a flag to remove all inherited annotations of the specified annotation type.
		Note that if this setting is enabled, it override all other settings.
	*/
	public void setRemove(boolean remove)
	{
		this.remove = remove;
	}
	
	public boolean getRemove()
	{
		return remove;
	}

	/**
		Set a flag to inherit annotations by cloning the original values. Use this
		if you do not want changes to the original annotation to be reflected in
		the inherited (=cloned) annotations.
		@since 3.6
	*/
	public void setClone(boolean clone)
	{
		this.clone = clone;
	}
	
	public boolean getClone()
	{
		return clone;
	}
	
	public void setResyncCloned(boolean resync)
	{
		this.resync = resync;
	}
	
	public boolean getResync()
	{
		return resync;
	}
	
	/**
		Set the subtype a parent item must have if annotations should be 
		inherited from it. Set to null to inherit from any parent.
	*/
	public void setItemSubtype(ItemSubtype subtype)
	{
		this.subtype = subtype;
	}
	
	public ItemSubtype getItemSubtype()
	{
		return subtype;
	}
	
	/**
		If this flag is set (default=true), only one annotation of the specified 
		annotation type can be inherited.
	*/
	public void setNoDuplicates(boolean noDuplicates)
	{
		this.noDuplicates = noDuplicates;
	}
	
	public boolean getNoDuplicates()
	{
		return noDuplicates;
	}
	
	/**
		If this flag is set (default=false), existing inherited annotations
		of the specified annotation type are removed before inheriting new 
		ones. Note that existing annotations are only replaced if a new annotation
		is found.
	*/
	public void setReplaceExisting(boolean replaceExisting)
	{
		this.replaceExisting = replaceExisting;
	}
	
	public boolean getReplaceExisting()
	{
		return replaceExisting;
	}

	/**
		Checks if the parent item has a subtype matching the subtype (if any)
		in this specification. If the parent item is a {@link BioPlate} we
		check the subtype of bioplate type ({@link BioPlateType#getItemSubtype()})
		@return TRUE if the subtype matches or if this specification doesn't set any subtype
	*/
	boolean isAcceptableParent(Annotatable parent)
	{
		if (subtype == null) return true;
		ItemSubtype subtypeToCheck = null;
		if (parent instanceof Subtypable)
		{
			subtypeToCheck = ((Subtypable)parent).getItemSubtype();
		}
		else if (parent instanceof BioPlate)
		{
			BioPlateType plateType = ((BioPlate)parent).getBioPlateType();
			subtypeToCheck = plateType != null ? plateType.getItemSubtype() : null;
		}
		return subtype.equals(subtypeToCheck);
	}
	
	
	@Override
	public boolean evaluate(AnnotationSnapshot snapshot) 
	{
		return snapshot.getAnnotationTypeId() == annotationTypeId;
	}


}
